@@ -3,18 +3,20 @@
 import datetime as dt
 import logging
 import random
-import select
+import selectors
 import signal
 import time
-from multiprocessing import Pool as MPool, cpu_count
+from multiprocessing import Pool as MPool
+from multiprocessing import cpu_count
 
 from sql import Flavor
 
 from trytond import backend
 from trytond.config import config
+from trytond.exceptions import UserError, UserWarning
 from trytond.pool import Pool
 from trytond.status import processing
-from trytond.transaction import Transaction
+from trytond.transaction import Transaction, TransactionError
 
 __all__ = ['work']
 logger = logging.getLogger(__name__)
@@ -62,27 +64,36 @@ def work(options):
     queues = [Queue(name, mpool) for name in options.database_names]
 
     tasks = TaskList()
+    selector = selectors.DefaultSelector()
+    for queue in queues:
+        selector.register(queue.connection, selectors.EVENT_READ)
     try:
         while True:
             timeout = options.timeout
+            # Add some randomness to avoid concurrent pulling
+            time.sleep(0.1 * random.random())
             while len(tasks.filter()) >= processes:
                 time.sleep(0.1)
             for queue in queues:
-                task_id, next_ = queue.pull(options.name)
+                try:
+                    task_id, next_ = queue.pull(options.name)
+                except backend.DatabaseOperationalError:
+                    break
                 if next_ is not None:
                     timeout = min(next_, timeout)
                 if task_id:
                     tasks.append(queue.run(task_id))
                     break
             else:
-                connections = [q.connection for q in queues]
-                connections, _, _ = select.select(connections, [], [], timeout)
-                for connection in connections:
+                for key, _ in selector.select(timeout=timeout):
+                    connection = key.fileobj
                     connection.poll()
                     while connection.notifies:
                         connection.notifies.pop(0)
     except KeyboardInterrupt:
         mpool.close()
+    finally:
+        selector.close()
 
 
 def initializer(database_names, worker=True):
@@ -107,14 +118,22 @@ def run_task(pool, task_id):
             with Transaction().start(pool.database_name, 0, readonly=True):
                 pool.init()
     Queue = pool.get('ir.queue')
+    Error = pool.get('ir.error')
+
+    def duration():
+        return (time.monotonic() - started) * 1000
+    started = time.monotonic()
     name = '<Task %s@%s>' % (task_id, pool.database_name)
-    logger.info('%s started', name)
     retry = config.getint('database', 'retry')
     try:
-        for count in range(retry, -1, -1):
-            if count != retry:
+        count = 0
+        transaction_extras = {}
+        while True:
+            if count:
                 time.sleep(0.02 * (retry - count))
-            with Transaction().start(pool.database_name, 0) as transaction:
+            with Transaction().start(
+                    pool.database_name, 0,
+                    **transaction_extras) as transaction:
                 try:
                     try:
                         task, = Queue.search([('id', '=', task_id)])
@@ -124,25 +143,47 @@ def run_task(pool, task_id):
                     with processing(name):
                         task.run()
                     break
+                except TransactionError as e:
+                    transaction.rollback()
+                    e.fix(transaction_extras)
+                    continue
                 except backend.DatabaseOperationalError:
-                    if count:
+                    if count < retry:
                         transaction.rollback()
+                        count += 1
+                        logger.debug("Retry: %i", count)
                         continue
                     raise
-        logger.info('%s done', name)
+                except (UserError, UserWarning) as e:
+                    Error.report(task, e)
+                    raise
+        logger.info("%s in %i ms", name, duration())
     except backend.DatabaseOperationalError:
-        logger.info('%s failed, retrying', name, exc_info=True)
+        logger.info(
+            "%s failed after %i ms, retrying", name, duration(),
+            exc_info=logger.isEnabledFor(logging.DEBUG))
         if not config.getboolean('queue', 'worker', default=False):
             time.sleep(0.02 * retry)
         try:
             with Transaction().start(pool.database_name, 0) as transaction:
+                if not transaction.database.has_channel():
+                    logger.critical('%s failed', name, exc_info=True)
+                    return
                 task = Queue(task_id)
-                scheduled_at = dt.datetime.now()
-                scheduled_at += dt.timedelta(
-                    seconds=random.randint(0, 2 * retry))
+                if task.scheduled_at and task.enqueued_at < task.scheduled_at:
+                    duration = (task.scheduled_at - task.enqueued_at) * 2
+                else:
+                    duration = dt.timedelta(seconds=2 * retry)
+                duration = max(duration, dt.timedelta(hours=1))
+                scheduled_at = dt.datetime.now() + duration * random.random()
                 Queue.push(task.name, task.data, scheduled_at=scheduled_at)
         except Exception:
             logger.critical(
-                'rescheduling %s failed', name, exc_info=True)
+                "rescheduling %s failed", name, exc_info=True)
+    except (UserError, UserWarning):
+        logger.info(
+            "%s failed after %i ms", name, duration(),
+            exc_info=logger.isEnabledFor(logging.DEBUG))
     except Exception:
-        logger.critical('%s failed', name, exc_info=True)
+        logger.critical(
+            "%s failed after %i ms", name, duration(), exc_info=True)
